@@ -4,7 +4,7 @@ use strict;
 use warnings;
 use parent qw/autobox/;
 
-our $VERSION = "1.002";
+our $VERSION = "1.003";
 
 =head1 NAME
 
@@ -14,7 +14,7 @@ autobox::Transform - Autobox methods to transform Arrays and Hashes
 
     # These are equivalent ways to transform arrays and arrayrefs
 
-    ### map_by
+    ### map_by - method call
     my @genres = map { $_->genre() } @$books;
     my @genres = $books->map_by("genre");
 
@@ -28,16 +28,24 @@ autobox::Transform - Autobox methods to transform Arrays and Hashes
     my $book_order_total = $order->books
         ->map_by(price_with_tax => [$tax_pct])->sum;
 
+    ### map_by - hash key
+    my @genres = map { $_->{genre} } @$books; # $books are hashrefs
+    my @genres = $books->map_by("genre");
 
-    ### grep_by
+
+    ### grep_by - method call
     my $sold_out_books = [ grep { $_->is_sold_out } @$books ];
     my $sold_out_books = $books->grep_by("is_sold_out");
 
     my $books_in_library = [ grep { $_->is_in_library($library) } @$books ];
     my $books_in_library = $books->grep_by(is_in_library => [$library]);
 
+    ### grep_by - hash key
+    my $sold_out_books = [ grep { $_->{is_sold_out} } @$books ]; # $books are hashrefs
+    my $sold_out_books = $books->grep_by("is_sold_out");
 
-    ### group_by
+
+    ### group_by - method call
 
     $books->group_by("title"),
     # {
@@ -66,6 +74,9 @@ autobox::Transform - Autobox methods to transform Arrays and Hashes
     #     "Fantasy" => [ $fantasy_book_1 ],
     # },
 
+    ### group_by - hash key
+    $books->group_by("title"), # $books are hashrefs
+
 
     #### flat
     my $prolific_author_books = [ map { @{$_->books} } @$authors ]
@@ -79,7 +90,27 @@ Note: This module supercedes autobox::Array::Transform which was
 unfortunately named.
 
 High level autobox methods you can call on arrays, arrayrefs, hashes
-and hashrefs e.g. map_by(), grep_by(), group_by()
+and hashrefs e.g.
+
+=over 4
+
+=item
+
+$array->map_by()
+
+=item
+
+$array->grep_by()
+
+=item
+
+$array->group_by()
+
+=item
+
+$array->flat()
+
+=back
 
 
 =head2 Raison d'etre
@@ -179,6 +210,18 @@ context. E.g.
     ),
 
 
+=head2 Transforming lists of objects vs list of hashrefs
+
+map_by, grep_by etc are called the same way regardless of whether the
+list contains objects or hashrefs. The items in the list must all be
+either objects or hashrefs.
+
+If an item is an object, the a method is called on it (possibly with
+the arguments provided).
+
+If it's a hashref, the hash key is looked up.
+
+
 
 =head1 AUTOBOX ARRAY METHODS
 
@@ -187,15 +230,25 @@ context. E.g.
 sub __invoke_by {
     my $invoke = shift;
     my $array = shift;
-    my( $method, $args ) = @_;
-    @_ > 0 or Carp::croak("->${invoke}_by() missing argument: \$method");
+    my( $accessor, $args ) = @_;
+    @_ > 0 or Carp::croak("->${invoke}_by() missing argument: \$accessor");
+    @$array or return wantarray ? () : [ ];
+
+    if ( ref($array->[0] ) eq "HASH" ) {
+        defined($args)
+            and Carp::croak("${invoke}_by('$accessor'): \$args ($args) only supported for method calls, not hash key access");
+        $invoke .= "_key";
+    }
+
     $args //= [];
     ref($args) eq "ARRAY"
-        or Carp::croak("${invoke}_by('$method', \$args): \$args ($args) is not an array ref");
+        or Carp::croak("${invoke}_by('$accessor', \$args): \$args ($args) is not an array ref");
 
     my $invoke_sub = {
-        map  => sub { [ CORE::map  { $_->$method( @$args ) } @$array ] },
-        grep => sub { [ CORE::grep { $_->$method( @$args ) } @$array ] },
+        map      => sub { [ CORE::map  { $_->$accessor( @$args ) } @$array ] },
+        grep     => sub { [ CORE::grep { $_->$accessor( @$args ) } @$array ] },
+        map_key  => sub { [ CORE::map  { $_->{$accessor}         } @$array ] },
+        grep_key => sub { [ CORE::grep { $_->{$accessor}         } @$array ] },
     }->{$invoke};
 
     my $result = eval { $invoke_sub->() }
@@ -204,11 +257,14 @@ sub __invoke_by {
     return wantarray ? @$result : $result;
 }
 
-=head2 map_by($method, @$args?) : @array | @$array
+=head2 map_by($accessor, @$args?) : @array | @$array
 
-Call the $method on each item in the list. Like:
+Call the $accessor on each object in the list, or get the hash key
+value on each hashref in the list. Like:
 
-    map { $_->$method() }
+    map { $_->$accessor() }
+    # or
+    map { $_->{$accessor} }
 
 Examples:
 
@@ -217,13 +273,17 @@ Examples:
 
 Optionally pass in @$args in the method call. Like:
 
-    map { $_->$method(@$args) }
+    map { $_->$accessor(@$args) }
 
 Examples:
 
     my @prices_including_tax = $books->map_by("price_with_tax", [ $tax_pct ]);
     my $prices_including_tax = $books->map_by(price_with_tax => [ $tax_pct ]);
 
+Or get the hash key value. Examples:
+
+    my @review_scores = $reviews->map_by("score");
+
 =cut
 
 sub map_by {
@@ -232,11 +292,13 @@ sub map_by {
 
 
 
-=head2 grep_by($method, @$args?) : @array | @$array
+=head2 grep_by($accessor, @$args?) : @array | @$array
 
-Call the $method on each item in the list. Like:
+Call the $accessor on each object in the list, or get the hash key
+value on each hashref in the list. Like:
 
-    grep { $_->$method() }
+    grep { $_->$accessor() }
+    grep { $_->{$accessor} }
 
 Examples:
 
@@ -244,7 +306,7 @@ Examples:
 
 Optionally pass in @$args in the method call. Like:
 
-    grep { $_->$method(@$args) }
+    grep { $_->$accessor(@$args) }
 
 Examples:
 
@@ -261,10 +323,11 @@ sub grep_by {
 
 
 
-=head2 group_by($method, @$args = [], $value_sub = object) : %key_value | %$key_value
+=head2 group_by($accessor, @$args = [], $value_sub = object) : %key_value | %$key_value
 
-Call ->$method(@$args) on each object in the array (just like ->map_by)
-and group the return values as keys in a hashref.
+Call ->$accessor(@$args) on each object in the array, or get the hash
+key for each hashref in the array (just like ->map_by) and group the
+return values as keys in a hashref.
 
 The default $value_sub puts the objects in the list as the hash
 values.
@@ -286,7 +349,7 @@ be to use one of the more specific group_by-methods which do common
 things (see below). It should be capable enough to achieve what you
 need though, so here's how it works:
 
-The hash key is whatever is returned from $object->$method(@$args).
+The hash key is whatever is returned from $object->$accessor(@$args).
 
 The hash value is whatever is returned from
 
@@ -306,28 +369,41 @@ $object is the current item in the list. The current $_ is also set to this.
 
 =item
 
-$key is the key returned by $object->$method(@$args)
+$key is the key returned by $object->$accessor(@$args)
 
 =back
 
 =cut
 
-sub group_by {
-    my $array = shift;
-    my( $method, $args, $value_sub ) = @_;
-    @_ > 0 or Carp::croak("->group_by() missing argument: \$method");
+sub __core_group_by {
+    my( $name, $array, $accessor, $args, $value_sub ) = @_;
+    $accessor or Carp::croak("->$name() missing argument: \$accessor");
+    @$array or return wantarray ? () : { };
 
-    $args //= [];
-    ref($args) eq "ARRAY"
-        or Carp::croak("group_by('$method', \$args, \$value_sub): \$args ($args) is not an array ref");
+    my $invoke = do {
+        # Hash key
+        if ( ref($array->[0] ) eq "HASH" ) {
+            defined($args)
+                and Carp::croak("$name('$accessor'): \$args ($args) only supported for method calls, not hash key access. Please specify an undef if needed.");
+            "key";
+        }
+        # Method
+        else {
+            $args //= [];
+            ref($args) eq "ARRAY"
+                or Carp::croak("$name('$accessor', \$args, \$value_sub): \$args ($args) is not an array ref");
+            "method";
+        }
+    };
 
-    $value_sub //= sub { $_ };
-    ref($value_sub) eq "CODE"
-        or Carp::croak("group_by('$method', [], \$value_sub): \$value_sub ($value_sub) is not a sub ref");
+    my $invoke_sub = {
+        method => sub { [ shift->$accessor(@$args) ] },
+        key    => sub { [ shift->{$accessor}       ] },
+    }->{$invoke};
 
     my %key_value;
     for my $object (@$array) {
-        my $key_ref = eval { [ $object->$method(@$args) ] }
+        my $key_ref = eval { $invoke_sub->($object) }
             or autobox::Transform::throw($@);
         my $key = $key_ref->[0];
 
@@ -341,10 +417,21 @@ sub group_by {
     return wantarray ? %key_value : \%key_value;
 }
 
-=head2 group_by_count($method, @$args = []) : %key_count | %$key_count
+sub group_by {
+    my $array = shift;
+    my( $accessor, $args, $value_sub ) = @_;
+
+    $value_sub //= sub { $_ };
+    ref($value_sub) eq "CODE"
+        or Carp::croak("group_by('$accessor', [], \$value_sub): \$value_sub ($value_sub) is not a sub ref");
+
+    return __core_group_by("group_by", $array, $accessor, $args, $value_sub);
+}
+
+=head2 group_by_count($accessor, @$args = []) : %key_count | %$key_count
 
 Just like group_by, but the hash values are the the number of
-instances each $method value occurs in the list.
+instances each $accessor value occurs in the list.
 
 Example:
 
@@ -361,25 +448,19 @@ for the "Sci-fi" key.
 
 sub group_by_count {
     my $array = shift;
-    my( $method, $args ) = @_;
-    ###JPL: extract args checking
-    @_ > 0 or Carp::croak("->group_by_count() missing argument: \$method");
-
-    $args //= [];
-    ref($args) eq "ARRAY"
-        or Carp::croak("group_by_count('$method', \$args): \$args ($args) is not an array ref");
+    my( $accessor, $args ) = @_;
 
     my $value_sub = sub {
         my $count = shift // 0; return ++$count;
     };
 
-    return group_by($array, $method, $args, $value_sub);
+    return __core_group_by("group_by_count", $array, $accessor, $args, $value_sub);
 }
 
-=head2 group_by_array($method, @$args = []) : %key_objects | %$key_objects
+=head2 group_by_array($accessor, @$args = []) : %key_objects | %$key_objects
 
 Just like group_by, but the hash values are arrayrefs containing the
-objects which has each $method value.
+objects which has each $accessor value.
 
 Example:
 
@@ -396,13 +477,7 @@ are collected under the Sci-fi key.
 
 sub group_by_array {
     my $array = shift;
-    my( $method, $args ) = @_;
-    ###JPL: extract args checking
-    @_ > 0 or Carp::croak("->group_by_array() missing argument: \$method");
-
-    $args //= [];
-    ref($args) eq "ARRAY"
-        or Carp::croak("group_by_array('$method', \$args): \$args ($args) is not an array ref");
+    my( $accessor, $args ) = @_;
 
     my $value_sub = sub {
         my $array = shift // [];
@@ -410,7 +485,7 @@ sub group_by_array {
         return $array;
     };
 
-    return group_by($array, $method, $args, $value_sub);
+    return __core_group_by("group_by_array", $array, $accessor, $args, $value_sub);
 }
 
 
@@ -449,20 +524,6 @@ sub flat {
 
 
 
-# TODO:
-# find(sub), like grep
-# find_by($method, $args, $value), like grep_by but first
-# contains value
-# compact - without undefined
-# compactTrue - without false
-# without - grep -v value
-
-
-# Hash::Transform: map_by() maps values from one thing to another
-
-
-
-
 =head1 DEVELOPMENT
 
 =head2 Author
